#!/usr/bin/python3

r"""
Plot simulated event distributions for shower and/or triggered event data.

Reads reduced event data files and generate histogram plots e.g. for energy or
core distance distributions.

Command line arguments
----------------------
input_file (str, required)
    Input file path.
output_path (str, required)
    Output directory for the generated plots.

Examples
--------
Generate plots from a given input file:

.. code-block:: console

    simtools-plot-simulated-event-distributions --input_file path/to/simtel_file.hdf5 \
                                                --output_path simtools_output/


"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.sim_events.histograms import EventDataHistograms
from simtools.visualization import plot_simtel_event_histograms


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Plot simulated event distributions for shower and/or triggered event data.",
    )
    config.parser.add_argument("--input_file", type=str, required=True, help="Input file path")
    return config.initialize(db_config=False, output=True)


def main():
    """Plot simulated event distributions."""
    app_context = startup_application(_parse)
    app_context.logger.info(f"Loading input file from: {app_context.args['input_file']}")

    histograms = EventDataHistograms(app_context.args["input_file"])
    histograms.fill()
    plot_simtel_event_histograms.plot(
        histograms.histograms, output_path=app_context.io_handler.get_output_directory()
    )


if __name__ == "__main__":
    main()



"""Helper functions for integration testing."""

from simtools import settings


def skip_camera_efficiency(config):
    """Skip camera efficiency tests if the old version of testeff is used."""
    if "camera-efficiency" in config["application"] and not _new_testeff_version():
        return (
            "Any applications calling the old version of testeff are skipped "
            "due to a limitation of the old testeff not allowing to specify "
            "the include directory. Please update your sim_telarray tarball."
        )
    return None


def _new_testeff_version():
    """
    Testeff has been updated to allow to specify the include directory.

    This test checks if the new version is used.
    """
    testeff_path = settings.config.sim_telarray_path / "testeff.c"
    try:
        with open(testeff_path, encoding="utf-8") as file:
            file_content = file.read()
            if (
                "/* Combine the include paths such that those from '-I...' options */"
                in file_content
            ):
                return True
            return False
    except FileNotFoundError as exc:
        raise FileNotFoundError("The testeff executable could not be found.") from exc


def skip_multiple_version_test(config, model_version):
    """Skip a test which is not meant for multiple versions if multiple versions are given."""
    message = "Skipping test not meant for multiple model versions."

    if not isinstance(model_version, list):
        return None

    config_model_version = config.get("configuration", {}).get("model_version", [])

    if not isinstance(config_model_version, list):
        config_model_version = [config_model_version]

    if 1 < len(model_version) != len(config_model_version):
        return message

    return None



"""Visualize Cherenkov photon distributions from CORSIKA."""

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from matplotlib import colormaps, colors

from simtools.visualization.visualize import save_figures_to_single_document


def _plot_2d(hist_list, labels=None):
    """
    Plot 2D Cherenkov photon distributions.

    Parameters
    ----------
    hist_list: list
        List of histogram dictionaries.
    labels: list or None
        Optional list of labels for the input files. If None, uses file names.

    Returns
    -------
    list
        List of figures.
    """
    if not hist_list:
        return []

    all_figs = []

    for i_file, hist_dict in enumerate(hist_list):
        hist_values = hist_dict["hist_values"]
        x_bin_edges = hist_dict["x_bin_edges"]
        y_bin_edges = hist_dict["y_bin_edges"]

        for i_hist, _ in enumerate(x_bin_edges):
            fig, ax = plt.subplots()
            if hist_dict.get("log_z", False):
                max_val = np.amax(hist_values[i_hist])
                norm = colors.LogNorm(vmin=1, vmax=np.amax([max_val, 2]))
            else:
                norm = None
            mesh = ax.pcolormesh(
                x_bin_edges[i_hist], y_bin_edges[i_hist], hist_values[i_hist], norm=norm
            )
            ax.set_xlabel(_get_axis_label(hist_dict["x_axis_title"], hist_dict["x_axis_unit"]))
            ax.set_ylabel(_get_axis_label(hist_dict["y_axis_title"], hist_dict["y_axis_unit"]))
            ax.set_xlim(np.amin(x_bin_edges[i_hist]), np.amax(x_bin_edges[i_hist]))
            ax.set_ylim(np.amin(y_bin_edges[i_hist]), np.amax(y_bin_edges[i_hist]))
            ax.set_facecolor("black")
            cbar = fig.colorbar(mesh)
            cbar.set_label(_get_axis_label(hist_dict["z_axis_title"], hist_dict["z_axis_unit"]))

            if labels is not None and i_file < len(labels):
                label = labels[i_file]
            else:
                label = Path(hist_dict.get("input_file_name", f"File {i_file}")).name
            ax.set_title(f"{hist_dict['title']} - {label}")

            all_figs.append(fig)
            plt.close()

    return all_figs


def _get_histogram_label(hist_dict, i_file, labels):
    """Get label for histogram curve."""
    if labels is not None and i_file < len(labels):
        return labels[i_file]
    return Path(hist_dict.get("input_file_name", f"File {i_file}")).name


def _extract_uncertainty(uncertainties, i_hist):
    """Extract uncertainty values if available."""
    if uncertainties is not None and uncertainties[i_hist] is not None:
        return uncertainties[i_hist]
    return None


def _plot_histogram_curve(ax, bin_centers, hist_values, uncertainties, color, label):
    """Plot a single histogram curve with or without error bars."""
    common_params = {
        "color": color,
        "label": label,
        "marker": "o",
        "markersize": 3,
        "linestyle": "-",
        "linewidth": 0.5,
    }

    if uncertainties is not None:
        ax.errorbar(
            bin_centers,
            hist_values,
            yerr=uncertainties,
            capsize=2,
            capthick=0.5,
            **common_params,
        )
    else:
        ax.plot(bin_centers, hist_values, **common_params)


def _configure_plot_scales(ax, hist):
    """Configure x and y axis scales."""
    if len(hist["x_bins"]) > 3 and hist["x_bins"][3] == "log":
        ax.set_xscale("log")
    if hist["log_y"] is True:
        ax.set_yscale("log")


def _plot_1d(hist_list, labels=None):
    """
    Plot 1D Cherenkov photon distributions.

    Parameters
    ----------
    hist_list: list
        List of histogram dictionaries from different files.
    labels: list or None
        Optional list of labels for the histogram curves. If None, uses file names.

    Returns
    -------
    list
        List of figures.
    """
    if not hist_list:
        return []

    hist = hist_list[0]
    plot_colors = colormaps["tab10"](np.linspace(0, 1, len(hist_list)))
    fig, ax = plt.subplots()

    for i_file, (hist_dict, color) in enumerate(zip(hist_list, plot_colors)):
        hist_values = hist_dict["hist_values"]
        x_bin_edges = hist_dict["x_bin_edges"]
        uncertainties = hist_dict.get("uncertainties")
        label = _get_histogram_label(hist_dict, i_file, labels)

        for i_hist, x_edges in enumerate(x_bin_edges):
            bin_centers = (x_edges[:-1] + x_edges[1:]) / 2
            unc = _extract_uncertainty(uncertainties, i_hist)
            _plot_histogram_curve(ax, bin_centers, hist_values[i_hist], unc, color, label)

    ax.set_xlabel(_get_axis_label(hist["x_axis_title"], hist["x_axis_unit"]))
    ax.set_ylabel(_get_axis_label(hist["y_axis_title"], hist["y_axis_unit"]))
    _configure_plot_scales(ax, hist)
    ax.set_title(f"{hist['title']}")
    ax.legend()
    ax.grid(True, alpha=0.3)

    plt.close(fig)
    return [fig]


def _get_axis_label(title, unit):
    """Return axis label with unit if applicable."""
    if unit is not u.dimensionless_unscaled:
        return f"{title} ({unit})"
    return f"{title}"


def _build_all_photon_figures(histograms_list, labels=None):
    """Build list of all photon histogram figures.

    Parameters
    ----------
    histograms_list: list
        List of CorsikaHistograms instances from different input files.
    labels: list or None
        Optional list of labels for the input files. If None, uses file names.

    Returns
    -------
    list
        List of figures.
    """
    all_figs = []

    if not isinstance(histograms_list, list):
        histograms_list = [histograms_list]

    hist_keys = list(histograms_list[0].hist.keys())

    for hist_key in hist_keys:
        hist_from_all_files = [h.hist[hist_key] for h in histograms_list]

        if hist_from_all_files[0]["is_1d"]:
            all_figs.extend(_plot_1d(hist_from_all_files, labels=labels))
        else:
            all_figs.extend(_plot_2d(hist_from_all_files, labels=labels))

    return all_figs


def export_all_photon_figures_pdf(histograms_instance, pdf_file_name, labels=None):
    """
    Build and save all photon histogram figures into a single PDF.

    Parameters
    ----------
    histograms_instance: corsika.corsika_histograms.CorsikaHistograms or list
        Single histogram instance or list of CorsikaHistograms instances from multiple files.
        When a list is provided, 1D histograms from all files are combined into single plots
        with different colors, while 2D histograms of the same type are plotted sequentially.
    pdf_file_name: str or Path
        Name of the output pdf file to save the histograms.
    labels: list or None
        Optional list of labels for the input files. If None, file names are used as labels.
        The order should match the order of histograms_instance if it's a list.
    """
    save_figures_to_single_document(
        _build_all_photon_figures(histograms_instance, labels=labels), Path(pdf_file_name)
    )



"""Helper functions for legend handlers used for plotting."""

# pylint: disable=too-few-public-methods

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import numpy as np

# Define properties of different telescope types for visualization purposes.
# Radii are relative to a reference radius (REFERENCE_RADIUS).
TELESCOPE_CONFIG = {
    "LST": {"color": "darkorange", "radius": 12.5, "shape": "circle", "filled": False},
    "MST": {"color": "dodgerblue", "radius": 9.15, "shape": "circle", "filled": False},
    "SCT": {"color": "black", "radius": 7.15, "shape": "square", "filled": False},
    "SST": {"color": "darkgreen", "radius": 3.0, "shape": "circle", "filled": False},
    "HESS": {"color": "grey", "radius": 6.0, "shape": "hexagon", "filled": True},
    "MAGIC": {"color": "grey", "radius": 8.5, "shape": "hexagon", "filled": True},
    "VERITAS": {"color": "grey", "radius": 6.0, "shape": "hexagon", "filled": True},
    "CEI": {"color": "purple", "radius": 2.0, "shape": "hexagon", "filled": True},
    "RLD": {"color": "brown", "radius": 2.0, "shape": "hexagon", "filled": True},
    "STP": {"color": "olive", "radius": 2.0, "shape": "hexagon", "filled": True},
    "MSP": {"color": "teal", "radius": 2.0, "shape": "hexagon", "filled": True},
    "ILL": {"color": "red", "radius": 2.0, "shape": "hexagon", "filled": False},
    "WST": {"color": "maroon", "radius": 2.0, "shape": "hexagon", "filled": True},
    "ASC": {"color": "cyan", "radius": 2.0, "shape": "hexagon", "filled": True},
    "DUS": {"color": "magenta", "radius": 2.0, "shape": "hexagon", "filled": True},
}

REFERENCE_RADIUS = 12.5


def get_telescope_config(telescope_type):
    """
    Return the configuration for a given telescope type.

    Try both site-dependent and site-independent configurations (e.g. "MSTS" and "MST").

    Parameters
    ----------
    telescope_type : str, None
        The type of the telescope (e.g., "LSTN", "MSTS").

    Returns
    -------
    dict
        The configuration dictionary for the telescope type.
    """
    if telescope_type is None:
        return {"color": "blue", "radius": 2.0, "shape": "hexagon", "filled": True}
    config = TELESCOPE_CONFIG.get(telescope_type)
    if not config and len(telescope_type) >= 3:
        config = TELESCOPE_CONFIG.get(telescope_type[:3])
    return config.copy() if config else None


def calculate_center(handlebox, width_factor=3, height_factor=3):
    """Calculate the center of the handlebox based on given factors."""
    x0 = handlebox.xdescent + handlebox.width / width_factor
    y0 = handlebox.ydescent + handlebox.height / height_factor
    return x0, y0


# Object classes for legend mapping
class PixelObject:
    """Pixel Object."""


class EdgePixelObject:
    """Edge-Pixel Object."""


class OffPixelObject:
    """Off-Pixel Object."""


class LSTObject:
    """LST Object."""


class MSTObject:
    """MST Object."""


class SCTObject:
    """SCT Object."""


class SSTObject:
    """SST Object."""


class HESSObject:
    """HESS Object."""


class MAGICObject:
    """MAGIC Object."""


class VERITASObject:
    """VERITAS Object."""


class MeanRadiusOuterEdgeObject:
    """Object for Mean radius outer edge."""


# Pixel handlers
class _BaseHexPixelHandler:
    """Base class for hexagonal pixel handlers."""

    @staticmethod
    def _create_hex_patch(handlebox, facecolor, edgecolor):
        """Create a hexagonal patch with specified colors."""
        x0, y0 = calculate_center(handlebox)
        patch = mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * handlebox.height,
            orientation=np.deg2rad(30),
            facecolor=facecolor,
            edgecolor=edgecolor,
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class HexPixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexPixelHandler._create_hex_patch(
            handlebox, facecolor=(1, 1, 1, 0), edgecolor=(0, 0, 0, 1)
        )


class HexEdgePixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexEdgePixelHandler._create_hex_patch(
            handlebox,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
        )


class HexOffPixelHandler(_BaseHexPixelHandler):
    """Legend handler class to plot a hexagonal "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return HexOffPixelHandler._create_hex_patch(handlebox, facecolor="black", edgecolor="black")


class _BaseSquarePixelHandler:
    """Base class for square pixel handlers."""

    @staticmethod
    def _create_square_patch(handlebox, facecolor, edgecolor):
        """Create a square patch with specified colors."""
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width = height = handlebox.height
        patch = mpatches.Rectangle(
            [x0, y0],
            width,
            height,
            facecolor=facecolor,
            edgecolor=edgecolor,
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


class SquarePixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "on" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquarePixelHandler._create_square_patch(
            handlebox, facecolor=(1, 1, 1, 0), edgecolor=(0, 0, 0, 1)
        )


class SquareEdgePixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "edge" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquareEdgePixelHandler._create_square_patch(
            handlebox,
            facecolor=(*mcolors.to_rgb("brown"), 0.5),
            edgecolor=(*mcolors.to_rgb("black"), 1),
        )


class SquareOffPixelHandler(_BaseSquarePixelHandler):
    """Legend handler class to plot a square "off" pixel."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):
        """Legend artist."""
        return SquareOffPixelHandler._create_square_patch(
            handlebox, facecolor="black", edgecolor="black"
        )


class BaseLegendHandler:
    """Base telescope handler that can handle any telescope type."""

    def __init__(self, telescope_type):
        self.telescope_type = telescope_type
        self.config = get_telescope_config(telescope_type)

    def _create_circle(self, handlebox, x0, y0, radius):
        """Create a circle patch."""
        facecolor = self.config["color"] if self.config["filled"] else "none"
        return mpatches.Circle(
            xy=(x0, y0),
            radius=radius * self.config["radius"] / REFERENCE_RADIUS,
            facecolor=facecolor,
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def _create_square(self, handlebox, x0, y0, size):
        """Create a square patch."""
        return mpatches.Rectangle(
            [x0, y0],
            size,
            size,
            facecolor=self.config["color"],
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def _create_hexagon(self, handlebox, x0, y0, radius):
        """Create a hexagon patch."""
        return mpatches.RegularPolygon(
            (x0, y0),
            numVertices=6,
            radius=0.7 * radius * self.config["radius"] / REFERENCE_RADIUS,
            orientation=np.deg2rad(30),
            facecolor=self.config["color"],
            edgecolor=self.config["color"],
            transform=handlebox.get_transform(),
        )

    def legend_artist(self, _, __, ___, handlebox):
        """Create the appropriate patch based on telescope type."""
        shape = self.config["shape"]

        if shape == "circle":
            x0, y0 = calculate_center(handlebox, 4, 2)
            radius = handlebox.height
            patch = self._create_circle(handlebox, x0, y0, radius)
        elif shape == "square":
            x0, y0 = calculate_center(handlebox, 10, 1)
            size = handlebox.height
            patch = self._create_square(handlebox, x0, y0, size)
        elif shape == "hexagon":
            x0, y0 = calculate_center(handlebox)
            radius = handlebox.height
            patch = self._create_hexagon(handlebox, x0, y0, radius)
        else:
            raise ValueError(f"Unknown shape: {shape}")

        handlebox.add_artist(patch)
        return patch


class LSTHandler(BaseLegendHandler):
    """Legend handler for LST telescopes."""

    def __init__(self):
        super().__init__("LST")


class MSTHandler(BaseLegendHandler):
    """Legend handler for MST telescopes."""

    def __init__(self):
        super().__init__("MST")


class SSTHandler(BaseLegendHandler):
    """Legend handler for SST telescopes."""

    def __init__(self):
        super().__init__("SST")


class SCTHandler(BaseLegendHandler):
    """Legend handler for SCT telescopes."""

    def __init__(self):
        super().__init__("SCT")


class HESSHandler(BaseLegendHandler):
    """Legend handler for HESS telescopes."""

    def __init__(self):
        super().__init__("HESS")


class MAGICHandler(BaseLegendHandler):
    """Legend handler for MAGIC telescopes."""

    def __init__(self):
        super().__init__("MAGIC")


class VERITASHandler(BaseLegendHandler):
    """Legend handler for VERITAS telescopes."""

    def __init__(self):
        super().__init__("VERITAS")


class MeanRadiusOuterEdgeHandler:
    """Legend handler class to plot the mean radius outer edge of the dish."""

    @staticmethod
    def legend_artist(_, __, ___, handlebox):  # noqa: D102
        x0, y0 = calculate_center(handlebox, 4, 4)
        radius = handlebox.height
        patch = mpatches.Circle(
            xy=(x0, y0),
            radius=radius,
            facecolor="none",
            edgecolor="darkorange",
            transform=handlebox.get_transform(),
        )
        handlebox.add_artist(patch)
        return patch


legend_handler_map = {
    LSTObject: LSTHandler,
    MSTObject: MSTHandler,
    SSTObject: SSTHandler,
    SCTObject: SCTHandler,
    HESSObject: HESSHandler,
    MAGICObject: MAGICHandler,
    VERITASObject: VERITASHandler,
}



"""Simulation runner for array simulations."""

import logging
import stat

from simtools import settings
from simtools.io import io_handler
from simtools.runners.simtel_runner import InvalidOutputFileError, SimtelRunner
from simtools.utils.general import clear_default_sim_telarray_cfg_directories


class SimulatorArray(SimtelRunner):
    """
    SimulatorArray is the interface with sim_telarray to perform array simulations.

    Parameters
    ----------
    corsika_config_data: CorsikaConfig
        CORSIKA configuration.
    label: str
        Instance label.
    use_multipipe: bool
        Use multipipe to run CORSIKA and sim_telarray.
    sim_telarray_seeds: dict
        Dictionary with configuration for sim_telarray random instrument setup.
    """

    def __init__(
        self,
        corsika_config,
        label=None,
        use_multipipe=False,
        sim_telarray_seeds=None,
        calibration_config=None,
    ):
        """Initialize SimulatorArray."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimulatorArray")
        super().__init__(
            label=label,
            corsika_config=corsika_config,
            use_multipipe=use_multipipe,
            calibration_run_mode=calibration_config.get("run_mode") if calibration_config else None,
        )

        self.sim_telarray_seeds = sim_telarray_seeds
        self.corsika_config = corsika_config
        self.calibration_config = calibration_config
        self.io_handler = io_handler.IOHandler()
        self._log_file = None

    def prepare_run_script(self, test=False, input_file=None, run_number=None, extra_commands=None):
        """
        Build and return the full path of the bash run script containing the sim_telarray command.

        Parameters
        ----------
        test: bool
            Test flag for faster execution.
        input_file: str or Path
            Full path of the input CORSIKA file.
        run_number: int
            Run number.
        extra_commands: list[str]
            Additional commands for running simulations given in config.yml.

        Returns
        -------
        Path
            Full path of the run script.
        """
        script_file_path = self.get_file_name(file_type="sub_script", run_number=run_number)
        self._logger.debug(f"Run bash script - {script_file_path}")
        self._logger.debug(f"Extra commands to be added to the run script {extra_commands}")

        command = self.make_run_command(run_number=run_number, input_file=input_file)
        with script_file_path.open("w", encoding="utf-8") as file:
            file.write("#!/usr/bin/env bash\n\n")
            file.write("set -e\n")
            file.write("set -o pipefail\n")
            file.write("\nSECONDS=0\n")

            if extra_commands:
                file.write("# Writing extras\n")
                for line in extra_commands:
                    file.write(f"{line}\n")
                file.write("# End of extras\n\n")

            n = 1 if test else self.runs_per_set
            for _ in range(n):
                file.write(f"{command}\n\n")

            file.write('\necho "RUNTIME: $SECONDS"\n')

        script_file_path.chmod(script_file_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP)
        return script_file_path

    def make_run_command(self, run_number=None, input_file=None, weak_pointing=None):
        """
        Build and return the command to run sim_telarray.

        Parameters
        ----------
        input_file: str
            Full path of the input CORSIKA file
        run_number: int (optional)
            run number
        weak_pointing: bool (optional)
            Specify weak pointing option for sim_telarray.

        Returns
        -------
        str
            Command to run sim_telarray.
        """
        command = self._common_run_command(run_number, weak_pointing)

        if self.calibration_config:
            command += self._make_run_command_for_calibration_simulations()
        else:
            command += self._make_run_command_for_shower_simulations()

        # "-C show=all" should be the last option
        command += super().get_config_option("show", "all")
        command += f" {input_file} | gzip > {self._log_file} 2>&1 || exit"

        return clear_default_sim_telarray_cfg_directories(command)

    def _make_run_command_for_shower_simulations(self):
        """
        Build and return the command to run sim_telarray shower simulations.

        Returns
        -------
        str
            Command to run sim_telarray.
        """
        return super().get_config_option(
            "power_law",
            SimulatorArray.get_power_law_for_sim_telarray_histograms(
                self.corsika_config.primary_particle
            ),
        )

    def _make_run_command_for_calibration_simulations(self):
        """Build sim_telarray command for calibration simulations."""
        cfg = self.calibration_config
        altitude = self.corsika_config.array_model.site_model.get_parameter_value_with_unit(
            "reference_point_altitude"
        ).to_value("m")

        command = super().get_config_option("Altitude", altitude)

        for key in ("nsb_scaling_factor", "stars"):
            if cfg.get(key):
                command += super().get_config_option(key, cfg[key])

        run_mode = cfg.get("run_mode")
        if run_mode in ("pedestals", "pedestals_nsb_only"):
            n_events = cfg.get("number_of_pedestal_events", cfg["number_of_events"])
            command += super().get_config_option("pedestal_events", n_events)
        if run_mode == "pedestals_nsb_only":
            command += self._pedestals_nsb_only_command()
        if run_mode == "pedestals_dark":
            n_events = cfg.get("number_of_dark_events", cfg["number_of_events"])
            command += super().get_config_option("dark_events", n_events)
        if run_mode == "direct_injection":
            n_events = cfg.get("number_of_flasher_events", cfg["number_of_events"])
            command += super().get_config_option("laser_events", n_events)

        return command

    def _common_run_command(self, run_number, weak_pointing=None):
        """Build generic run command for sim_telarray."""
        config_dir = self.corsika_config.array_model.get_config_directory()
        self._log_file = self.get_file_name(file_type="log", run_number=run_number)
        histogram_file = self.get_file_name(file_type="histogram", run_number=run_number)
        output_file = self.get_file_name(file_type="simtel_output", run_number=run_number)
        self.corsika_config.array_model.export_all_simtel_config_files()

        command = str(settings.config.sim_telarray_exe)
        command += f" -c {self.corsika_config.array_model.config_file_path}"
        command += f" -I{config_dir}"
        command += super().get_config_option(
            "telescope_theta", self.corsika_config.zenith_angle, weak_pointing
        )
        command += super().get_config_option(
            "telescope_phi", self.corsika_config.azimuth_angle, weak_pointing
        )
        command += super().get_config_option("histogram_file", histogram_file)
        command += super().get_config_option("random_state", "none")
        if self.sim_telarray_seeds and self.sim_telarray_seeds.get("random_instrument_instances"):
            command += super().get_config_option(
                "random_seed",
                f"file-by-run:{config_dir}/{self.sim_telarray_seeds['seed_file_name']},auto",
            )
        elif self.sim_telarray_seeds and self.sim_telarray_seeds.get("seed"):
            command += super().get_config_option("random_seed", self.sim_telarray_seeds["seed"])
        command += super().get_config_option("output_file", output_file)

        return command

    def _pedestals_nsb_only_command(self):
        """
        Generate the command to run sim_telarray for nsb-only pedestal simulations.

        Returns
        -------
        str
            Command to run sim_telarray.
        """
        null_values = [
            "fadc_noise",
            "fadc_lg_noise",
            "qe_variation",
            "gain_variation",
            "fadc_var_pedestal",
            "fadc_err_pedestal",
            "fadc_sysvar_pedestal",
            "fadc_dev_pedestal",
        ]
        null_command_parts = [super().get_config_option(param, 0.0) for param in null_values]
        command = " ".join(null_command_parts)

        one_values = [
            "fadc_lg_var_pedestal",
            "fadc_lg_err_pedestal",
            "fadc_lg_dev_pedestal",
            "fadc_lg_sysvar_pedestal",
        ]
        one_command_parts = [super().get_config_option(param, -1.0) for param in one_values]
        command += " " + " ".join(one_command_parts)
        return command

    def _check_run_result(self, run_number=None):
        """
        Check if sim_telarray output file exists.

        Parameters
        ----------
        run_number: int
            Run number.

        Returns
        -------
        bool
            True if sim_telarray output file exists.

        Raises
        ------
        InvalidOutputFileError
            If sim_telarray output file does not exist.
        """
        output_file = self.get_file_name(file_type="simtel_output", run_number=run_number)
        if not output_file.exists():
            raise InvalidOutputFileError(f"sim_telarray output file {output_file} does not exist.")
        self._logger.debug(f"sim_telarray output file {output_file} exists.")
        return True

    @staticmethod
    def get_power_law_for_sim_telarray_histograms(primary):
        """
        Get the power law index for sim_telarray.

        Events will be filled in histograms in sim_telarray with a weight according to
        the difference between this exponent and the one used for the shower simulations.

        Parameters
        ----------
        primary: str
            Primary particle.

        Returns
        -------
        float
            Power law index.
        """
        power_laws = {
            "gamma": 2.5,
            "electron": 3.3,
        }
        if primary.name in power_laws:
            return power_laws[primary.name]

        return 2.68



"""
Simtools dependencies version management.

This modules provides two main functionalities:

- retrieve the versions of simtools dependencies (e.g., databases, sim_telarray, CORSIKA)
- provide space for future implementations of version management

"""

import logging
import re
import subprocess
from pathlib import Path

import yaml

from simtools import settings
from simtools.io import ascii_handler
from simtools.utils import general as gen
from simtools.version import __version__

_logger = logging.getLogger(__name__)


def get_version_string(run_time=None):
    """
    Print the versions of the dependencies.

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        String containing the versions of the dependencies.

    """
    return (
        f"Database name: {get_database_version_or_name(version=False)}\n"
        f"Database version: {get_database_version_or_name(version=True)}\n"
        f"sim_telarray version: {get_sim_telarray_version(run_time)}\n"
        "sim_telarray exe: "
        f"{settings.config.sim_telarray_exe if settings.config.sim_telarray_exe else 'None'}\n"
        f"CORSIKA version: {get_corsika_version(run_time)}\n"
        f"CORSIKA exe: {settings.config.corsika_exe if settings.config.corsika_exe else 'None'}\n"
        f"Build options: {get_build_options(run_time)}\n"
        f"Runtime environment: {run_time if run_time else 'None'}\n"
    )


def get_software_version(software):
    """
    Return the version of the specified software package.

    Parameters
    ----------
    software : str
        Name of the software package.

    Returns
    -------
    str
        Version of the specified software package.
    """
    if software.lower() == "simtools":
        return __version__

    try:
        version_call = f"get_{software.lower()}_version"
        return globals()[version_call]()
    except KeyError as exc:
        raise ValueError(f"Unknown software: {software}") from exc


def get_database_version_or_name(version=True):
    """
    Get the version or name of the simulation model data base used.

    Parameters
    ----------
    version : bool
        If True, return the version of the database. If False, return the name.

    Returns
    -------
    str
        Version or name of the simulation model data base used.

    """
    if version:
        return settings.config.db_config and settings.config.db_config.get(
            "db_simulation_model_version"
        )
    return settings.config.db_config and settings.config.db_config.get("db_simulation_model")


def get_sim_telarray_version(run_time=None):
    """
    Get the version of the sim_telarray package using 'sim_telarray --version'.

    Version strings for sim_telarray are of the form "2024.271.0" (year.day_of_year.patch).

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        Version of the sim_telarray package.
    """
    if settings.config.sim_telarray_exe is None:
        _logger.warning("sim_telarray environment not configured.")
        return None
    if run_time is None:
        command = [str(settings.config.sim_telarray_exe), "--version"]
    else:
        command = [*run_time, str(settings.config.sim_telarray_exe), "--version"]

    _logger.debug(f"Running command: {command}")
    result = subprocess.run(command, capture_output=True, text=True, check=False)

    # expect stdout with e.g. a line 'Release: 2024.271.0 from 2024-09-27'
    match = re.search(r"^Release:\s+(.+)", result.stdout, re.MULTILINE)
    if match:
        return match.group(1).split()[0]

    _logger.debug(f"Command output stdout: {result.stdout} stderr: {result.stderr}")

    raise ValueError(f"sim_telarray release not found in {result.stdout}")


def get_corsika_version(run_time=None):
    """
    Get the version of the CORSIKA package.

    Version strings for CORSIKA are of the form "7.7550" (major.minor with 4-digit minor).

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    str
        Version of the CORSIKA package.
    """
    if settings.config.corsika_exe is None:
        _logger.warning("CORSIKA environment not configured.")
        return None

    if run_time is None:
        command = [str(settings.config.corsika_exe)]
    else:
        command = [*run_time, str(settings.config.corsika_exe)]

    process = subprocess.Popen(  # pylint: disable=consider-using-with
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        text=True,
    )

    version = None
    for line in process.stdout:
        # Extract the version from the line "NUMBER OF VERSION :  7.7550"
        if "NUMBER OF VERSION" in line:
            version = line.split(":")[1].strip()
            break
        # Check for a specific prompt or indication that the program is waiting for input
        if "DATA CARDS FOR RUN STEERING ARE EXPECTED FROM STANDARD INPUT" in line:
            break

    process.terminate()
    # Check it's a valid version string
    if version and re.match(r"\d+\.\d+", version):
        return version
    try:
        build_opts = get_build_options(run_time)
    except (FileNotFoundError, TypeError, ValueError):
        _logger.warning("Could not get CORSIKA version.")
        return None
    _logger.debug("Getting the CORSIKA version from the build options.")
    return build_opts.get("corsika_version")


def get_build_options(run_time=None):
    """
    Return CORSIKA / sim_telarray config and build options.

    For CORSIKA / sim_telarray build for simtools version >0.25.0:
    expects build_opts.yml file in each CORSIKA and sim_telarray
    directories.

    For CORSIKA / sim_telarray build for simtools version <=0.25.0:
    expects a build_opts.yml file in the sim_telarray directory.

    Parameters
    ----------
    run_time : list, optional
        Runtime environment command (e.g., Docker).

    Returns
    -------
    dict
        CORSIKA / sim_telarray build options.
    """
    build_opts = {}
    for package in ["corsika", "sim_telarray"]:
        path = _get_package_path(package)
        if not path:
            continue
        try:
            build_opts.update(_get_build_options_from_file(path / "build_opts.yml", run_time))
        except (FileNotFoundError, TypeError, ValueError):
            # legacy fallback only for sim_telarray
            if package == "sim_telarray":
                try:
                    legacy_path = path.parent / "build_opts.yml"
                    build_opts.update(_get_build_options_from_file(legacy_path, run_time))
                except (FileNotFoundError, TypeError, ValueError):
                    _logger.debug(f"No build options found for {package}.")
    if not build_opts:
        raise FileNotFoundError("No build option file found.")

    return build_opts


def _get_package_path(package):
    """Get the package path from settings or environment variables."""
    path = getattr(settings.config, f"{package}_path")
    if path is None:
        path = gen.load_environment_variables().get(f"{package}_path")
    return Path(path) if path else None


def _get_build_options_from_file(build_opts_path, run_time=None):
    """Read build options from file."""
    if run_time is None:
        try:
            return ascii_handler.collect_data_from_file(build_opts_path)
        except FileNotFoundError as exc:
            raise FileNotFoundError("No build option file found.") from exc

    command = [*run_time, "cat", str(build_opts_path)]
    _logger.debug(f"Reading build option with command: {command}")

    result = subprocess.run(command, capture_output=True, text=True, check=False)
    if result.returncode:
        raise FileNotFoundError(f"No build option file found in container: {result.stderr}")

    try:
        return yaml.safe_load(result.stdout)
    except yaml.YAMLError as exc:
        raise ValueError(f"Error parsing build_opts.yml from container: {exc}") from exc


def export_build_info(output_file, run_time=None):
    """
    Export build and version information to a file.

    Parameters
    ----------
    output_file : str
        Path to the output file.
    run_time : list, optional
        Runtime environment command (e.g., Docker).
    """
    build_info = get_build_options(run_time)
    build_info["simtools"] = __version__
    build_info["database_name"] = get_database_version_or_name(version=False)
    build_info["database_version"] = get_database_version_or_name(version=True)
    ascii_handler.write_data_to_file(data=build_info, output_file=Path(output_file))



#!/usr/bin/python3
"""
Handling of legacy model parameters for backward compatibility.

Collects routines to convert legacy model parameters to more recent
formats. This is a fine-tuned process and requires to hard wire the
changes. All code related to legacy model parameters should go into
this module.

"""

import logging

logger = logging.getLogger(__name__)


UPDATE_HANDLERS = {}


def register_update(name):
    """Register update handler for legacy model parameter."""

    def deco(func):
        UPDATE_HANDLERS[name] = func
        return func

    return deco


def apply_legacy_updates_to_parameters(parameters, legacy_updates):
    """Apply legacy updates to model parameters.

    Modifies the parameters dictionary in-place.

    Parameters
    ----------
    parameters: dict
        Dictionary of model parameters (all parameters).
    legacy_updates: dict
        Dictionary of legacy updates to apply.
    """
    for par_name, legacy_data in legacy_updates.items():
        if legacy_data is None or par_name not in parameters:
            continue
        for key in parameters[par_name].keys():
            if key in legacy_data:
                parameters[par_name][key] = legacy_data[key]
        if legacy_data.get("remove_parameter", False):
            parameters.pop(par_name)


def update_parameter(par_name, parameters, schema_version):
    """Update legacy model parameters to recent formats.

    Parameters
    ----------
    par_name: str
        Model parameter name.
    parameters: dict
        Dictionary of model parameters (all parameters).
    schema_version: str
        Target schema version.

    Returns
    -------
    dict
        Updated model parameter.
    """
    handler = UPDATE_HANDLERS.get(par_name)
    if handler is None:
        raise ValueError(_get_unsupported_update_message(parameters[par_name], schema_version))
    return handler(parameters, schema_version)


@register_update("dsum_threshold")
def _update_dsum_threshold(parameters, schema_version):
    """Update legacy dsum_threshold parameter."""
    para_data = parameters["dsum_threshold"]
    if para_data["model_parameter_schema_version"] == "0.1.0" and schema_version == "0.2.0":
        logger.info(
            "Updating legacy model parameter dsum_threshold from schema version "
            f"{para_data['model_parameter_schema_version']} to {schema_version}"
        )
        return {
            para_data["parameter"]: {
                "value": int(para_data["value"]),
                "model_parameter_schema_version": schema_version,
            }
        }
    raise ValueError(_get_unsupported_update_message(para_data, schema_version))


@register_update("corsika_starting_grammage")
def _update_corsika_starting_grammage(parameters, schema_version):
    """Update legacy corsika_starting_grammage parameter (dummy function until model is updated)."""
    logger.debug(f"No fix applied to corsika_starting_grammage to schema version {schema_version}")
    return {
        parameters["corsika_starting_grammage"]["parameter"]: None,
    }


@register_update("flasher_pulse_shape")
def _update_flasher_pulse_shape(parameters, schema_version):
    """Update legacy flasher_pulse_shape parameter."""
    para_data = parameters["flasher_pulse_shape"]
    if para_data["model_parameter_schema_version"] == "0.1.0" and schema_version == "0.2.0":
        logger.info(
            f"Updating legacy model parameter flasher_pulse_shape from schema version "
            f"{para_data['model_parameter_schema_version']} to {schema_version}"
        )
        return {
            para_data["parameter"]: {
                "value": [
                    para_data["value"],
                    parameters.get("flasher_pulse_width", {}).get("value", 0.0),
                    parameters.get("flasher_pulse_exp_decay", {}).get("value", 0.0),
                ],
                "model_parameter_schema_version": schema_version,
                "unit": [None, "ns", "ns"],
                "type": ["string", "float64", "float64"],
            },
            "flasher_pulse_width": {"remove_parameter": True},
            "flasher_pulse_exp_decay": {"remove_parameter": True},
        }

    raise ValueError(_get_unsupported_update_message(para_data, schema_version))


def _get_unsupported_update_message(para_data, schema_version):
    """Get unsupported update message."""
    return (
        f"Unsupported update for legacy parameter {para_data['parameter']}: "
        f"{para_data['model_parameter_schema_version']} to {schema_version}"
    )



"""Extract Cherenkov photons from a CORSIKA IACT file and fill histograms."""

import functools
import logging
import operator
from pathlib import Path

import boost_histogram as bh
import numpy as np
from astropy import units as u
from eventio import IACTFile

from simtools.utils.geometry import rotate


class CorsikaHistograms:
    """
    Extract Cherenkov photons from a CORSIKA IACT file and fill histograms.

    Parameters
    ----------
    input_file: str or Path
        CORSIKA IACT file.

    Raises
    ------
    FileNotFoundError:
        if the input file given does not exist.
    """

    def __init__(self, input_file):
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CorsikaHistograms")
        self.input_file = Path(input_file)
        if not self.input_file.exists():
            raise FileNotFoundError(f"File {self.input_file} does not exist.")

        self.events = None
        self.hist = self._set_2d_distributions()
        self.hist.update(self._set_1d_distributions())

    def fill(self):
        """
        Fill Cherenkov photons histograms.

        Returns
        -------
        list: list of boost_histogram.Histogram instances.

        Raises
        ------
        AttributeError:
            if event has not photon saved.
        """
        self._read_event_headers()

        with IACTFile(self.input_file) as f:
            telescope_positions = np.array(f.telescope_positions)
            for event_counter, event in enumerate(f):
                if hasattr(event, "photon_bunches"):
                    photons = list(event.photon_bunches.values())
                    self._fill_histograms(photons, event_counter, telescope_positions, True)

        self._update_distributions()

    def _read_event_headers(self):
        """Read event information from headers."""
        event_dtype = np.dtype(
            [
                ("particle_id", "i4"),
                ("total_energy", "f8"),
                ("azimuth_deg", "f8"),
                ("zenith_deg", "f8"),
                ("num_photons", "f8"),
            ]
        )

        with IACTFile(self.input_file) as iact_file:
            records = [
                (
                    event.header["particle_id"],
                    event.header["total_energy"],
                    np.rad2deg(event.header["azimuth"]),
                    np.rad2deg(event.header["zenith"]),
                    0.0,  # filled later when reading photon bunches
                )
                for event in iact_file
            ]

        self.events = np.array(records, dtype=event_dtype)

    def _create_regular_axes(self, hist, axes):
        """
        Create regular axis for a single histogram.

        Parameters
        ----------
        hist: dict
            Histogram dictionary.
        axes: list
            List of axis names (e.g. ["x_bins", "y_bins"]).

        Returns
        -------
        list:
            List of boost_histogram axis instances.
        """
        transform = {"log": bh.axis.transform.log, "linear": None}

        boost_axes = []
        for axis in axes:
            bins, start, stop = hist[axis][:3]
            scale = hist[axis][3] if len(hist[axis]) > 3 else "linear"
            if isinstance(start, u.quantity.Quantity):
                start, stop = start.value, stop.value
            boost_axes.append(
                bh.axis.Regular(
                    bins=bins,
                    start=start,
                    stop=stop,
                    transform=transform[scale],
                )
            )
        return boost_axes

    def _fill_histograms(self, photons, event_counter, telescope_positions, rotate_photons=True):
        """
        Fill Cherenkov photon histograms.

        For rotate_photons, the Cherenkov photon's coordinates are filled in the shower plane.

        Parameters
        ----------
        photons: list
            List of size M of numpy.array of size (N,8), where M is the number of telescopes in the
            array, N is the number of photons that reached each telescope. The following information
             of the Cherenkov photons on the ground are saved:
             x: x position on the ground (CORSIKA coordinate system),
             y: y position on the ground (CORSIKA coordinate system),
             cx: direction cosine in the x direction, i.e., the cosine of the angle between the
             incoming direction and the x axis,
             cy: direction cosine in the y direction, i.e., the cosine of the angle between the
             incoming direction and the y axis,
             time: time of arrival of the photon in ns.
             zem: altitude where the photon was generated in cm,
             photons: number of photons associated to this bunch,
        event_counter: int
            Event counter.
        telescope_positions: numpy.array
            Array with the telescope positions with shape (M, 2), where M is the number of
            telescopes in the array. The two columns are the x and y positions of the telescopes
            in the CORSIKA coordinate system.
        rotate_photons: bool
            If True, the photon's coordinates are rotated to the plane perpendicular to the
            incoming direction of the primary particle.
        """
        hist_str = "histogram"
        for photon, telescope in zip(photons, telescope_positions):
            if rotate_photons:
                px, py = rotate(
                    photon["x"],
                    photon["y"],
                    self.events["azimuth_deg"][event_counter],
                    self.events["zenith_deg"][event_counter],
                )
            else:
                px, py = photon["x"], photon["y"]

            px -= telescope["x"]
            py -= telescope["y"]
            w = photon["photons"]

            pxm = px * u.cm.to(u.m)
            pym = py * u.cm.to(u.m)
            zem = (photon["zem"] * u.cm).to(u.km)

            self.hist["counts_xy"][hist_str].fill(pxm, pym, weight=w)
            self.hist["density_xy"][hist_str].fill(pxm, pym, weight=w)
            self.hist["direction_xy"][hist_str].fill(photon["cx"], photon["cy"], weight=w)
            self.hist["time_altitude"][hist_str].fill(photon["time"] * u.ns, zem, weight=w)
            self.hist["wavelength_altitude"][hist_str].fill(
                np.abs(photon["wavelength"]) * u.nm, zem, weight=w
            )

            r = np.hypot(px, py) * u.cm.to(u.m)
            self.hist["counts_r"][hist_str].fill(r, weight=w)
            self.hist["density_r"][hist_str].fill(r, weight=w)

            self.events["num_photons"][event_counter] += np.sum(w)

    def get_hist_2d_projection(self, hist):
        """
        Get 2D distributions.

        Parameters
        ----------
        hist: boost_histogram.Histogram
            Histogram.

        Returns
        -------
        numpy.ndarray
            Histogram counts.
        numpy.array
            Histogram x bin edges.
        numpy.array
            Histogram y bin edges.
        numpy.ndarray or None
            Histogram uncertainties (sqrt of variance) if available.
        """
        view = hist.view()
        if self._check_for_all_attributes(view):
            counts = np.asarray([view["value"].T])
            uncertainties = np.asarray([np.sqrt(view["variance"].T)])
        else:
            counts = np.asarray([view.T])
            uncertainties = None

        x_edges = np.asarray([hist.axes.edges[0].flatten()])
        y_edges = np.asarray([hist.axes.edges[1].flatten()])

        return counts, x_edges, y_edges, uncertainties

    def _get_hist_1d_from_numpy(self, label, hist):
        """Get 1D histogram from numpy histogram."""
        bins = hist["x_bins"][0]
        start = hist["x_bins"][1] if hist["x_bins"][1] else np.min(self.events[label])
        stop = hist["x_bins"][2] if hist["x_bins"][2] is not None else np.max(self.events[label])
        scale = hist["x_bins"][3] if len(hist["x_bins"]) > 3 else "linear"
        if scale == "log":
            bin_edges = np.logspace(np.log10(start), np.log10(stop), bins + 1)
        else:
            bin_edges = np.linspace(start, stop, bins + 1)
        histo_1d, _ = np.histogram(self.events[label], bins=bin_edges)
        uncertainties = np.sqrt(histo_1d)
        return (
            histo_1d.reshape(1, bins),
            bin_edges.reshape(1, bins + 1),
            uncertainties.reshape(1, bins),
        )

    def get_hist_1d_projection(self, label, hist):
        """
        Get 1D distributions from numpy or boost histograms (1D and 2D).

        Parameters
        ----------
        label: str
            Histogram label.
        hist: dict
            Histogram dictionary.

        Returns
        -------
        numpy.ndarray
            Histogram counts.
        numpy.array
            Histogram x bin edges.
        numpy.ndarray or None
            Histogram uncertainties (if available).
        """
        # plain numpy histogram
        if (
            hist.get("projection") is None
            and hasattr(self, "events")
            and label in self.events.dtype.names
        ):
            return self._get_hist_1d_from_numpy(label, hist)

        # boost 1D histogram
        if hist.get("projection") is None:
            # Use histogram from hist dict if available, otherwise from self.hist
            if "histogram" in hist:
                histo_1d = hist["histogram"]
            else:
                # No histogram available, return None values
                return None, None, None
            edges = histo_1d.axes.edges.T.flatten()[0]
            view = histo_1d.view()
            if self._check_for_all_attributes(view):
                counts = np.asarray([view["value"].T])
                uncertainties = np.asarray([np.sqrt(view["variance"].T)])
            else:
                counts = np.asarray([view.T])
                uncertainties = None
            return counts, np.asarray([edges]), uncertainties

        # boost 2D histogram projection
        histo_2d = self.hist[hist["projection"][0]]["histogram"]
        if hist["projection"][1] == "x":
            h = histo_2d[:, sum]
        else:
            h = histo_2d[sum, :]
        edges = h.axes.edges.T.flatten()[0]
        view = h.view()
        if self._check_for_all_attributes(view):
            counts = np.asarray([view["value"].T])
            uncertainties = np.asarray([np.sqrt(view["variance"].T)])
        else:
            counts = np.asarray([view.T])
            uncertainties = None
        return counts, np.asarray([edges]), uncertainties

    def _set_1d_distributions(self, r_max=2000 * u.m, bins=100):
        """
        Define 1D histograms.

        Returns
        -------
        dict:
            Dictionary with 1D histogram information.
        """
        file_name = "file_name"
        title = "title"
        projection = "projection"
        x_bins = "x_bins"
        x_axis_unit = "x_axis_unit"
        x_axis_title = "x_axis_title"
        y_axis_unit = "y_axis_unit"
        y_axis_title = "y_axis_title"
        hist_1d = {
            "wavelength": {
                file_name: "hist_1d_photon_wavelength_distr",
                title: "Photon wavelength distribution",
                projection: ["wavelength_altitude", "x"],
            },
            "counts_r": {
                file_name: "hist_1d_photon_radial_distr",
                title: "Photon lateral distribution (ground level)",
                x_bins: [bins, 0 * u.m, r_max, "linear"],
                x_axis_title: "Distance to center",
                x_axis_unit: u.m,
            },
            "density_r": {
                file_name: "hist_1d_photon_density_distr",
                title: "Photon lateral density distribution (ground level)",
                x_bins: [bins, 0 * u.m, r_max, "linear"],
                x_axis_title: "Distance to center",
                x_axis_unit: u.m,
                y_axis_title: "Photon density",
                y_axis_unit: u.m**-2,
            },
            "density_x": {
                file_name: "hist_1d_photon_density_x_distr",
                title: "Photon lateral density x distribution (ground level)",
                projection: ["density_xy", "x"],
            },
            "density_y": {
                file_name: "hist_1d_photon_density_y_distr",
                title: "Photon lateral density y distribution (ground level)",
                projection: ["density_xy", "y"],
            },
            "time": {
                file_name: "hist_1d_photon_time_distr",
                title: "Photon arrival time distribution",
                projection: ["time_altitude", "x"],
            },
            "altitude": {
                file_name: "hist_1d_photon_altitude_distr",
                title: "Photon emission altitude distribution",
                projection: ["time_altitude", "y"],
            },
            "direction_cosine_x": {
                file_name: "hist_1d_photon_direction_cosine_x_distr",
                title: "Photon direction cosine x distribution",
                projection: ["direction_xy", "x"],
            },
            "direction_cosine_y": {
                file_name: "hist_1d_photon_direction_cosine_y_distr",
                title: "Photon direction cosine y distribution",
                projection: ["direction_xy", "y"],
            },
            "num_photons": {
                file_name: "hist_1d_photon_per_event_distr",
                title: "Photons per event distribution",
                "event_type": True,
                x_bins: [100, 0, None, "log"],
                x_axis_title: "Cherenkov photons per event",
                x_axis_unit: u.dimensionless_unscaled,
            },
        }

        for value in hist_1d.values():
            value["is_1d"] = True
            value["log_y"] = True
            value[y_axis_title] = (
                "Counts" if value.get(y_axis_title) is None else value[y_axis_title]
            )
            value[y_axis_unit] = (
                u.dimensionless_unscaled if value.get(y_axis_unit) is None else value[y_axis_unit]
            )
            if value.get("projection") is not None:
                hist_2d_name = value["projection"][0]
                if value["projection"][1] == "x":
                    value[x_bins] = self.hist[hist_2d_name]["x_bins"]
                    value[x_axis_title] = self.hist[hist_2d_name]["x_axis_title"]
                    value[x_axis_unit] = self.hist[hist_2d_name]["x_axis_unit"]
                else:
                    value[x_bins] = self.hist[hist_2d_name]["y_bins"]
                    value[x_axis_title] = self.hist[hist_2d_name]["y_axis_title"]
                    value[x_axis_unit] = self.hist[hist_2d_name]["y_axis_unit"]
            elif value.get("event_type", False) is False:
                boost_axes = self._create_regular_axes(value, ["x_bins"])
                value["histogram"] = bh.Histogram(boost_axes[0], storage=bh.storage.Weight())
        return hist_1d

    def _set_2d_distributions(self, xy_maximum=1000 * u.m, xy_bin=100):
        """
        Define 2D histograms.

        Returns
        -------
        dict:
            Dictionary with 2D histogram information.
        """
        file_name = "file_name"
        title = "title"
        x_bins, y_bins = "x_bins", "y_bins"
        x_axis_title, x_axis_unit = "x_axis_title", "x_axis_unit"
        y_axis_title, y_axis_unit = "y_axis_title", "y_axis_unit"
        z_axis_title, z_axis_unit = "z_axis_title", "z_axis_unit"

        hist_2d = {
            "counts_xy": {
                file_name: "hist_2d_photon_count_distr",
                title: "Photon count distribution (ground level)",
                x_bins: [xy_bin, -xy_maximum, xy_maximum, "linear"],
                y_bins: [xy_bin, -xy_maximum, xy_maximum],
                x_axis_title: "x position on the ground",
                x_axis_unit: xy_maximum.unit,
                y_axis_title: "y position on the ground",
                y_axis_unit: xy_maximum.unit,
            },
            "density_xy": {
                file_name: "hist_2d_photon_density_distr",
                title: "Photon lateral density distribution (ground level)",
                x_bins: [xy_bin, -xy_maximum, xy_maximum, "linear"],
                y_bins: [xy_bin, -xy_maximum, xy_maximum, "linear"],
                x_axis_title: "x position on the ground",
                x_axis_unit: xy_maximum.unit,
                y_axis_title: "y position on the ground",
                y_axis_unit: xy_maximum.unit,
                z_axis_title: "Photon density",
                z_axis_unit: u.m**-2,
            },
            "direction_xy": {
                file_name: "hist_2d_photon_direction_distr",
                title: "Photon arrival direction",
                x_bins: [100, -1, 1, "linear"],
                y_bins: [100, -1, 1, "linear"],
                x_axis_title: "x direction cosine",
                x_axis_unit: u.dimensionless_unscaled,
                y_axis_title: "y direction cosine",
                y_axis_unit: u.dimensionless_unscaled,
            },
            "time_altitude": {
                file_name: "hist_2d_photon_time_altitude_distr",
                title: "Arrival time vs emission altitude",
                x_bins: [100, -2000 * u.ns, 2000 * u.ns, "linear"],
                y_bins: [100, 120 * u.km, 0 * u.km, "linear"],
                x_axis_title: "Arrival time",
                x_axis_unit: u.ns,
                y_axis_title: "Emission altitude",
                y_axis_unit: u.km,
            },
            "wavelength_altitude": {
                file_name: "hist_2d_photon_wavelength_altitude_distr",
                title: "Wavelength vs emission altitude ",
                x_bins: [100, 100 * u.nm, 1000 * u.nm, "linear"],
                y_bins: [100, 120 * u.km, 0 * u.km, "linear"],
                x_axis_title: "Wavelength",
                x_axis_unit: u.nm,
                y_axis_title: "Emission altitude",
                y_axis_unit: u.km,
            },
        }

        for value in hist_2d.values():
            value["is_1d"] = False
            value["log_z"] = True
            value[z_axis_title] = (
                "Counts" if value.get(z_axis_title) is None else value[z_axis_title]
            )
            value[z_axis_unit] = (
                u.dimensionless_unscaled if value.get(z_axis_unit) is None else value[z_axis_unit]
            )
            boost_axes = self._create_regular_axes(value, ["x_bins", "y_bins"])
            value["histogram"] = bh.Histogram(
                boost_axes[0], boost_axes[1], storage=bh.storage.Weight()
            )

        return hist_2d

    def _update_distributions(self):
        """Update the distributions dictionary with the histogram values and bin edges."""
        self._normalize_density_histograms()

        for key, value in self.hist.items():
            value["input_file_name"] = str(self.input_file)
            if value["is_1d"]:
                value["hist_values"], value["x_bin_edges"], value["uncertainties"] = (
                    self.get_hist_1d_projection(key, value)
                )
            else:
                (
                    value["hist_values"],
                    value["x_bin_edges"],
                    value["y_bin_edges"],
                    value["uncertainties"],
                ) = self.get_hist_2d_projection(value["histogram"])

    def _normalize_density_histograms(self):
        """Normalize the density histograms by the area of each bin."""

        def normalize_histogram(hist, bin_areas):
            view = hist.view()
            if self._check_for_all_attributes(view):
                view["value"] /= bin_areas
                view["variance"] /= bin_areas**2
            else:
                hist /= bin_areas

        def normalize_histogram_1d(hist, bin_areas):
            view = hist.view()
            if self._check_for_all_attributes(view):
                view["value"] /= bin_areas
                view["variance"] /= bin_areas**2
            else:
                view /= bin_areas

        density_xy_hist = self.hist["density_xy"]["histogram"]
        density_r_hist = self.hist["density_r"]["histogram"]

        bin_areas_xy = functools.reduce(operator.mul, density_xy_hist.axes.widths)
        normalize_histogram(density_xy_hist, bin_areas_xy)

        bin_edges_r = density_r_hist.axes.edges[0]
        bin_areas_r = np.pi * (bin_edges_r[1:] ** 2 - bin_edges_r[:-1] ** 2)
        normalize_histogram_1d(density_r_hist, bin_areas_r)

    def _check_for_all_attributes(self, view):
        """Check if view has dtype fields ('value', 'variance')."""
        if hasattr(view, "dtype") and view.dtype.names == ("value", "variance"):
            return True
        return False



#!/usr/bin/python3

"""
Make a regular array of telescopes and save it as astropy table.

The arrays consist of one telescope at the center of the array and or of 4 telescopes
in a square grid. These arrays are used for trigger rate simulations.

The array layout files created will be available at the data/layout directory.

Command line arguments
----------------------
site (str, required)
    observatory site (e.g., North or South).
model_version (str, optional)
    Model version to use (e.g., 6.0.0). If not provided, the latest version is used.

Example
-------
Runtime < 10 s.

.. code-block:: console

    simtools-generate-regular-arrays --site=North
"""

from pathlib import Path

import astropy.units as u

import simtools.data_model.model_data_writer as writer
from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.layout.array_layout_utils import create_regular_array

# Telescope distances for 4 tel square arrays
# !HARDCODED
telescope_distance = {"LST": 57.5 * u.m, "MST": 70 * u.m, "SST": 80 * u.m}


def _parse():
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description=(
            "Generate a regular array of telescope and save as astropy table.\n"
            "Default telescope distances for 4 telescope square arrays are: \n"
            f"  LST: {telescope_distance['LST']}\n"
            f"  MST: {telescope_distance['MST']}\n"
            f"  SST: {telescope_distance['SST']}\n"
        ),
    )
    return config.initialize(
        db_config=False, simulation_model=["site", "model_version"], output=True
    )


def main():
    """Create layout array files (ecsv) of regular arrays."""
    app_context = startup_application(_parse)

    if app_context.args["site"] == "South":
        array_list = ["1SST", "4SST", "1MST", "4MST", "1LST", "4LST"]
    else:
        array_list = ["1MST", "4MST", "1LST", "4LST"]

    for array_name in array_list:
        app_context.logger.info(f"Processing array {array_name}")

        array_table = create_regular_array(
            array_name, app_context.args["site"], telescope_distance=telescope_distance
        )

        output_file = app_context.args.get("output_file")
        if output_file:
            output_path = Path(output_file)
            output_file = output_path.with_name(
                f"{output_path.stem}-{app_context.args['site']}-{array_name}{output_path.suffix}"
            )
        writer.ModelDataWriter.dump(
            args_dict=app_context.args,
            output_file=output_file,
            metadata=None,
            product_data=array_table,
        )


if __name__ == "__main__":
    main()



"""Centralized settings object with command line and environment variables."""

import os
from pathlib import Path
from types import MappingProxyType


class _Config:
    """Centralized settings object with command line and environment variables."""

    def __init__(self):
        """Initialize empty config."""
        self._args = {}
        self._db_config = {}
        self._sim_telarray_path = None
        self._sim_telarray_exe = None
        self._corsika_path = None
        self._corsika_exe = None

    def load(self, args=None, db_config=None):
        """
        Load configuration from command line arguments and environment variables.

        For paths, first check for environment variables, then command line arguments.

        Parameters
        ----------
        args : dict, optional
            Command line arguments.
        db_config : dict, optional
            Database configuration.

        """
        self._args = MappingProxyType(args) if args is not None else {}
        self._db_config = MappingProxyType(db_config) if db_config is not None else {}
        self._sim_telarray_path = (
            args.get("sim_telarray_path")
            if args is not None and "sim_telarray_path" in args
            else os.getenv("SIMTOOLS_SIM_TELARRAY_PATH")
        )

        self._sim_telarray_exe = (
            args.get("sim_telarray_executable")
            if args is not None and "sim_telarray_executable" in args
            else os.getenv("SIMTOOLS_SIM_TELARRAY_EXECUTABLE", "sim_telarray")
        )

        self._corsika_path = (
            args.get("corsika_path")
            if args is not None and "corsika_path" in args
            else os.getenv("SIMTOOLS_CORSIKA_PATH")
        )

        self._corsika_exe = self._get_corsika_exec() if self._corsika_path is not None else None

    def _get_corsika_exec(self):
        """
        Get the CORSIKA executable from environment variable or command line argument.

        Build the executable name based on configured interaction models. Fall back to
        legacy naming (simply "corsika") if models are not specified.
        """
        he_model = (
            self._args.get("corsika_he_interaction")
            if self._args is not None and "corsika_he_interaction" in self._args
            else os.getenv("SIMTOOLS_CORSIKA_HE_INTERACTION")
        )

        le_model = (
            self._args.get("corsika_le_interaction")
            if self._args is not None and "corsika_le_interaction" in self._args
            else os.getenv("SIMTOOLS_CORSIKA_LE_INTERACTION")
        )

        if he_model and le_model:
            corsika_exe = self.corsika_path / f"corsika_{he_model}_{le_model}_flat"
            if corsika_exe.exists():
                return corsika_exe

        # legacy naming
        return self.corsika_path / "corsika"

    @property
    def args(self):
        """Command line arguments."""
        return self._args

    @property
    def db_config(self):
        """Database configuration."""
        return self._db_config

    @property
    def sim_telarray_path(self):
        """Path to the sim_telarray installation directory."""
        return Path(self._sim_telarray_path) if self._sim_telarray_path is not None else None

    @property
    def sim_telarray_exe(self):
        """Path to the sim_telarray executable."""
        return (
            Path(self._sim_telarray_path) / "bin" / self._sim_telarray_exe
            if self._sim_telarray_path is not None
            else None
        )

    @property
    def sim_telarray_exe_debug_trace(self):
        """Path to the debug trace version of the sim_telarray executable."""
        return (
            Path(self._sim_telarray_path) / "bin" / (self._sim_telarray_exe + "_debug_trace")
            if self._sim_telarray_path is not None
            else None
        )

    @property
    def corsika_path(self):
        """Path to the CORSIKA installation directory."""
        return Path(self._corsika_path) if self._corsika_path is not None else None

    @property
    def corsika_exe(self):
        """Path to the CORSIKA executable."""
        return (
            Path(self._corsika_path) / self._corsika_exe if self._corsika_path is not None else None
        )

    @property
    def corsika_exe_curved(self):
        """Path to the curved version of the CORSIKA executable."""
        if self._corsika_exe is None:
            return None
        corsika_curved = (
            self._corsika_exe.name.replace("_flat", "_curved")
            if "_flat" in self._corsika_exe.name
            else self._corsika_exe.name + "-curved"  # legacy naming convention
        )
        return Path(self._corsika_path) / corsika_curved if self._corsika_path is not None else None

    @property
    def corsika_dummy_file(self):
        """
        Path to a dummy CORSIKA file required by sim_telarray for ray-tracing simulations.

        This file does not need to exist; sim_telarray only requires a file path.
        """
        return (
            self.sim_telarray_path / "run9991.corsika.gz"
            if self._sim_telarray_path is not None
            else None
        )


config = _Config()



"""Base class for running sim_telarray simulations."""

import logging
import subprocess

import simtools.utils.general as gen
from simtools.runners.runner_services import RunnerServices


class SimtelExecutionError(Exception):
    """Exception for sim_telarray execution error."""


class InvalidOutputFileError(Exception):
    """Exception for invalid output file."""


class SimtelRunner:
    """
    Base class for running simulations based on the sim_telarray software stack.

    The sim_telarray software stack includes sim_telarray itself and e.g., testeff,
    LightEmission, and other software packages.

    Parameters
    ----------
    label: str
        Instance label. Important for output file naming.
    corsika_config: CorsikaConfig
        CORSIKA configuration.
    use_multipipe: bool
        Use multipipe to run CORSIKA and sim_telarray.
    calibration_run_mode: str
        Calibration run mode.
    """

    def __init__(
        self,
        label=None,
        corsika_config=None,
        use_multipipe=False,
        calibration_run_mode=None,
    ):
        """Initialize SimtelRunner."""
        self._logger = logging.getLogger(__name__)

        self.label = label
        self._base_directory = None
        self.calibration_run_mode = calibration_run_mode

        self.runs_per_set = 1

        self.runner_service = RunnerServices(corsika_config, label)
        self._directory = self.runner_service.load_data_directories(
            "corsika_sim_telarray" if use_multipipe else "sim_telarray"
        )

    def run(self, test=False, input_file=None, run_number=None):
        """
        Make run command and run sim_telarray.

        Parameters
        ----------
        test: bool
            If True, make simulations faster.
        input_file: str or Path
            Full path of the input CORSIKA file.
        run_number: int
            Run number.
        """
        self._logger.debug("Running sim_telarray")

        command, stdout_file, stderr_file = self._make_run_command(
            run_number=run_number, input_file=input_file
        )

        if test:
            self._logger.info(f"Running (test) with command: {command}")
            self._run_simtel_and_check_output(command, stdout_file, stderr_file)
        else:
            self._logger.debug(f"Running ({self.runs_per_set}x) with command: {command}")
            for _ in range(self.runs_per_set):
                self._run_simtel_and_check_output(command, stdout_file, stderr_file)

        self._check_run_result(run_number=run_number)

    def _check_run_result(self, run_number=None):  # pylint: disable=all
        """Check if simtel output file exists."""
        pass

    def _raise_simtel_error(self):
        """
        Raise sim_telarray execution error.

        Final 30 lines from the log file are collected and printed.

        Raises
        ------
        SimtelExecutionError
        """
        if hasattr(self, "_log_file"):
            msg = gen.get_log_excerpt(self._log_file)
        else:
            msg = "Simtel log file does not exist."
        raise SimtelExecutionError(msg)

    def _run_simtel_and_check_output(self, command, stdout_file, stderr_file):
        """
        Run the sim_telarray command and check the exit code.

        Raises
        ------
        SimtelExecutionError
            if run was not successful.
        """
        stdout_file = stdout_file if stdout_file else "/dev/null"
        stderr_file = stderr_file if stderr_file else "/dev/null"
        with (
            open(f"{stdout_file}", "w", encoding="utf-8") as stdout,
            open(f"{stderr_file}", "w", encoding="utf-8") as stderr,
        ):
            result = subprocess.run(
                command,
                shell=True,
                text=True,
                stdout=stdout,
                stderr=stderr,
            )

        if result.returncode != 0:
            self._logger.error(result.stderr)
            self._raise_simtel_error()
        return result.returncode

    def _make_run_command(self, run_number=None, input_file=None):
        self._logger.debug(
            "make_run_command is being called from the base class - "
            "it should be implemented in the sub class"
        )
        input_file = input_file if input_file else "nofile"
        run_number = run_number if run_number else 1
        return f"{input_file}-{run_number}", None, None

    @staticmethod
    def get_config_option(par, value=None, weak_option=False):
        """
        Build sim_telarray command.

        Parameters
        ----------
        par: str
            Parameter name.
        value: str
            Parameter value.
        weak_option: bool
            If True, use -W option instead of -C.

        Returns
        -------
        str
            Command for sim_telarray.
        """
        option_syntax = "-W" if weak_option else "-C"
        c = f" {option_syntax} {par}"
        c += f"={value}" if value is not None else ""
        return c

    def get_resources(self, run_number=None):
        """Return computing resources used."""
        return self.runner_service.get_resources(run_number)

    def get_file_name(
        self,
        simulation_software="sim_telarray",
        file_type=None,
        run_number=None,
        mode="",
        model_version_index=0,
    ):
        """
        Get the full path of a file for a given run number.

        Parameters
        ----------
        simulation_software: str
            Simulation software.
        file_type: str
            File type.
        run_number: int
            Run number.
        model_version_index: int
            Index of the model version.
            This is used to select the correct simulator_array instance in case
            multiple array models are simulated.

        Returns
        -------
        str
            File name with full path.
        """
        if simulation_software.lower() != "sim_telarray":
            raise ValueError(
                f"simulation_software ({simulation_software}) is not supported in SimulatorArray"
            )
        return self.runner_service.get_file_name(
            file_type=file_type,
            run_number=run_number,
            mode=mode,
            _model_version_index=model_version_index,
            calibration_run_mode=self.calibration_run_mode,
        )



#!/usr/bin/python3

r"""
    Generates a set of histograms with Cherenkov photon distributions from CORSIKA output.

    The Cherenkov photons (from observation level) are read from a CORSIKA IACT
    output file(s) provided as input.

    The following 2D histograms are generated:

        - Density of Cherenkov photons on the ground
        - Incoming direction (directive cosines) of the Cherenkov photons
        - Time of arrival (ns) vs altitude of production (km)

    The following 1D histograms are generated:

        - Wavelength distribution of Cherenkov photons
        - Time of arrival (ns) distribution of Cherenkov photons
        - Altitude of production (km) distribution of Cherenkov photons
        - Lateral distribution of Cherenkov photons (distance from shower core in m)

    Command line arguments
    ----------------------
    input_files (str, required)
        The name(s) of the CORSIKA IACT file(s) resulted from the CORSIKA simulation.

    pdf_file_name (str, optional)
        The name of the output pdf file to save the histograms. If not provided,
        the histograms are only shown on screen.

    file_labels (str, optional)
        Labels for the input files (in the same order as input_files). If not provided,
        the file names are used as labels.

    Example
    -------
    Fill and plot histograms for a test IACT file:

     .. code-block:: console

        simtools-generate-corsika-histograms --input_files /workdir/external/simtools/\\
        tests/resources/tel_output_10GeV-2-gamma-20deg-CTAO-South.corsikaio \\
            --pdf_file_name test.pdf

    Fill and plot histograms for several files:

     .. code-block:: console

        simtools-generate-corsika-histograms --input_files file1 file 2 \\
            --file_lablels label1 label2 \\
            --pdf_file_name test.pdf

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.corsika.corsika_histograms import CorsikaHistograms
from simtools.visualization import plot_corsika_histograms


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Generate histograms for the Cherenkov photons saved in the CORSIKA IACT file.",
    )
    config.parser.add_argument(
        "--input_files",
        help="Name(s) of the CORSIKA IACT file(s) to process",
        type=str,
        nargs="+",
        required=True,
    )
    config.parser.add_argument(
        "--file_labels",
        help="Labels for the input files (in the same order as input_files)",
        type=str,
        nargs="+",
        required=None,
    )
    config.parser.add_argument(
        "--pdf_file_name",
        help="Save histograms into a pdf file.",
        type=str,
        required=None,
    )
    return config.initialize(db_config=False, paths=True)


def main():
    """Generate a set of histograms for the Cherenkov photons from CORSIKA IACT file(s)."""
    app_context = startup_application(_parse)

    all_histograms = []
    for input_file in app_context.args["input_files"]:
        corsika_histograms = CorsikaHistograms(input_file)
        corsika_histograms.fill()
        all_histograms.append(corsika_histograms)

    plot_corsika_histograms.export_all_photon_figures_pdf(
        all_histograms,
        app_context.io_handler.get_output_file(app_context.args.get("pdf_file_name")),
        app_context.args.get("file_labels"),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Plot array elements for a layout."""

from collections import Counter
from typing import NamedTuple

import astropy.units as u
import matplotlib as mpl
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from adjustText import adjust_text
from astropy.table import Column
from matplotlib.collections import PatchCollection

from simtools.utils import geometry as transf
from simtools.utils import names
from simtools.visualization import legend_handlers as leg_h
from simtools.visualization import visualize


class PlotBounds(NamedTuple):
    """Axis-aligned bounds for the layout in meters.

    Attributes
    ----------
    x_lim : tuple[float, float]
        Min/max for x (meters).
    y_lim : tuple[float, float]
        Min/max for y (meters).
    """

    x_lim: tuple[float, float]
    y_lim: tuple[float, float]


def plot_array_layouts(args_dict, output_path, layouts, background_layout=None):
    """
    Plot multiple array layouts.

    Parameters
    ----------
    args_dict : dict
        Application arguments.
    output_path : Path
        Output path for figures.
    layouts : dict
        Dictionary of layout name to telescope table.
    background_layout : Table or None
        Optional background telescope table.

    Returns
    -------
    figs : dict
        Dictionary of layout name to matplotlib figure object.

    """
    mpl.use("Agg")
    for layout in layouts:
        fig_out = plot_array_layout(
            telescopes=layout["array_elements"],
            show_tel_label=args_dict["show_labels"],
            axes_range=args_dict["axes_range"],
            marker_scaling=args_dict["marker_scaling"],
            background_telescopes=background_layout,
            grayed_out_elements=args_dict["grayed_out_array_elements"],
            highlighted_elements=args_dict["highlighted_array_elements"],
            legend_location=args_dict["legend_location"],
            bounds_mode=args_dict["bounds"],
            padding=args_dict["padding"],
            x_lim=tuple(args_dict["x_lim"]) if args_dict["x_lim"] else None,
            y_lim=tuple(args_dict["y_lim"]) if args_dict["y_lim"] else None,
        )
        site_string = ""
        if layout.get("site") is not None:
            site_string = f"_{layout['site']}"
        elif args_dict["site"] is not None:
            site_string = f"_{args_dict['site']}"
        coordinate_system_string = (
            f"_{args_dict['coordinate_system']}"
            if args_dict["coordinate_system"] not in layout["name"]
            else ""
        )
        plot_file_name = args_dict["figure_name"] or (
            f"array_layout_{layout['name']}{site_string}{coordinate_system_string}"
        )

        visualize.save_figure(fig_out, output_path / plot_file_name, dpi=400)
        plt.close()


def plot_array_layout(
    telescopes,
    show_tel_label=False,
    axes_range=None,
    marker_scaling=1.0,
    background_telescopes=None,
    grayed_out_elements=None,
    highlighted_elements=None,
    legend_location="best",
    bounds_mode="exact",
    padding=0.1,
    x_lim=None,
    y_lim=None,
):
    """
    Plot telescope array layout.

    Parameters
    ----------
    telescopes : Table
        Telescope data table.
    show_tel_label : bool
        Show telescope labels (default False).
    axes_range : float or None
        Axis range, auto if None.
    marker_scaling : float
        Marker size scale factor.
    background_telescopes : Table or None
        Optional background telescope table.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.
    legend_location : str
        Location of the legend (default "best").

    Returns
    -------
    fig : Figure
        Matplotlib figure object.

    Other Parameters
    ----------------
    bounds_mode : {"symmetric", "exact"}
        Controls axis limits calculation. "symmetric" uses +-R where R is the padded
        maximum extent (default), while "exact" uses individual x/y min/max bounds.
    padding : float
        Fractional padding applied around computed extents (used for both modes).
    x_lim, y_lim : tuple(float, float), optional
        Explicit axis limits in meters. If provided, these override axes_range and bounds_mode
        for the respective axis. If only one is provided, the other axis is derived per mode.
    """
    fig, ax = plt.subplots(1)

    # If explicit limits are provided (one or both), filter patches accordingly
    filter_x = x_lim
    filter_y = y_lim

    patches, plot_range, highlighted_patches, bounds, text_objects = get_patches(
        ax,
        telescopes,
        show_tel_label,
        axes_range,
        marker_scaling,
        grayed_out_elements,
        highlighted_elements,
        filter_x_lim=filter_x,
        filter_y_lim=filter_y,
    )

    plot_range, bounds = _get_patches_for_background_telescopes(
        ax,
        background_telescopes,
        axes_range,
        marker_scaling,
        bounds_mode,
        plot_range,
        bounds,
        filter_x_lim=filter_x,
        filter_y_lim=filter_y,
    )

    if legend_location != "no_legend":
        update_legend(ax, telescopes, grayed_out_elements, legend_location)

    x_lim, y_lim = _get_axis_limits(
        axes_range, bounds_mode, padding, plot_range, bounds, x_lim, y_lim
    )

    finalize_plot(ax, patches, "Easting [m]", "Northing [m]", x_lim, y_lim, highlighted_patches)

    if text_objects:
        adjust_text(
            text_objects,
            ax=ax,
            arrowprops={"arrowstyle": "->", "color": "grey", "alpha": 0.8, "lw": 0.8, "ls": "--"},
            expand=(2.0, 2.0),
            prevent_crossings=True,
            min_arrow_len=8,
            ensure_inside_axes=True,
        )

    return fig


def _get_axis_limits(
    axes_range,
    bounds_mode,
    padding,
    plot_range,
    bounds,
    x_lim_override=None,
    y_lim_override=None,
):
    """Get axis limits based on mode and padding."""

    def _derive_axis(axis: str) -> tuple[float, float]:
        if bounds_mode == "exact":
            if axis == "x":
                span = bounds.x_lim[1] - bounds.x_lim[0]
                pad = padding * span
                return (bounds.x_lim[0] - pad, bounds.x_lim[1] + pad)
            span = bounds.y_lim[1] - bounds.y_lim[0]
            pad = padding * span
            return (bounds.y_lim[0] - pad, bounds.y_lim[1] + pad)
        # symmetric
        sym = plot_range
        padf = max(0.0, min(1.0, float(padding))) if padding is not None else 0.0
        sym *= 1.0 + padf
        return (-sym, sym)

    # Highest priority: explicit overrides (per axis)
    if x_lim_override is not None or y_lim_override is not None:
        x_lim = x_lim_override if x_lim_override is not None else _derive_axis("x")
        y_lim = y_lim_override if y_lim_override is not None else _derive_axis("y")
        return x_lim, y_lim

    if axes_range is not None:
        return (-axes_range, axes_range), (-axes_range, axes_range)
    # Derive both axes using selected mode
    return _derive_axis("x"), _derive_axis("y")


def _get_patches_for_background_telescopes(
    ax,
    background_telescopes,
    axes_range,
    marker_scaling,
    bounds_mode,
    plot_range,
    bounds,
    filter_x_lim=None,
    filter_y_lim=None,
):
    """Get background telescope patches and update plot range/bounds."""
    if background_telescopes is None:
        return plot_range, bounds

    bg_patches, bg_range, _, bg_bounds, _ = get_patches(
        ax,
        background_telescopes,
        False,
        axes_range,
        marker_scaling,
        None,
        None,
        filter_x_lim=filter_x_lim,
        filter_y_lim=filter_y_lim,
    )
    ax.add_collection(PatchCollection(bg_patches, match_original=True, alpha=0.1))
    if axes_range is None:
        if bounds_mode == "symmetric":
            plot_range = max(plot_range, bg_range)
        else:
            bounds = PlotBounds(
                x_lim=(
                    min(bounds.x_lim[0], bg_bounds.x_lim[0]),
                    max(bounds.x_lim[1], bg_bounds.x_lim[1]),
                ),
                y_lim=(
                    min(bounds.y_lim[0], bg_bounds.y_lim[0]),
                    max(bounds.y_lim[1], bg_bounds.y_lim[1]),
                ),
            )
    return plot_range, bounds


def _apply_limits_filter(telescopes, pos_x, pos_y, filter_x_lim, filter_y_lim):
    """Filter telescope table and positions by optional axis limits."""
    if filter_x_lim is None and filter_y_lim is None:
        return telescopes, pos_x, pos_y

    px = np.asarray(pos_x.to_value(u.m))
    py = np.asarray(pos_y.to_value(u.m))
    mask = np.ones(px.shape, dtype=bool)
    if filter_x_lim is not None:
        mask &= (px >= float(filter_x_lim[0])) & (px <= float(filter_x_lim[1]))
    if filter_y_lim is not None:
        mask &= (py >= float(filter_y_lim[0])) & (py <= float(filter_y_lim[1]))

    if mask.size and mask.any():
        return telescopes[mask], pos_x[mask], pos_y[mask]

    # No telescopes within limits
    return telescopes[:0], pos_x[:0], pos_y[:0]


def get_patches(
    ax,
    telescopes,
    show_tel_label,
    axes_range,
    marker_scaling,
    grayed_out_elements=None,
    highlighted_elements=None,
    filter_x_lim=None,
    filter_y_lim=None,
):
    """
    Get plot patches and axis range.

    Parameters
    ----------
    ax : Axes
        Matplotlib axes object.
    telescopes : Table
        Telescope data table.
    show_tel_label : bool
        Show telescope labels.
    axes_range : float or None
        Axis range, auto if None.
    marker_scaling : float
        Marker size scale factor.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.

    Returns
    -------
    patches : list
        List of telescope patches.
    axes_range : float
        Calculated or input symmetric axis range (meters).
    highlighted_patches : list
        List of highlighted telescope patches.
    bounds : PlotBounds
        Min/max for x and y in meters.
    text_objects : list
        List of text objects for labels.
    """
    pos_x, pos_y = get_positions(telescopes)
    tel_table, pos_x, pos_y = _apply_limits_filter(
        telescopes, pos_x, pos_y, filter_x_lim, filter_y_lim
    )

    tel_table["pos_x_rotated"] = Column(pos_x)
    tel_table["pos_y_rotated"] = Column(pos_y)

    patches, radii, highlighted_patches, text_objects = create_patches(
        tel_table, marker_scaling, show_tel_label, ax, grayed_out_elements, highlighted_elements
    )

    if len(radii) == 0:
        r = 0.0
    else:
        radii_q = u.Quantity(radii)
        r = float(np.nanmax(radii_q).to_value(u.m))

    if len(pos_x) == 0:
        bounds = PlotBounds(x_lim=(0.0, 0.0), y_lim=(0.0, 0.0))
        if axes_range:
            return patches, axes_range, highlighted_patches, bounds, text_objects
        return patches, 0.0, highlighted_patches, bounds, text_objects

    x_min = float(np.nanmin(pos_x).to_value(u.m)) - r
    x_max = float(np.nanmax(pos_x).to_value(u.m)) + r
    y_min = float(np.nanmin(pos_y).to_value(u.m)) - r
    y_max = float(np.nanmax(pos_y).to_value(u.m)) + r
    bounds = PlotBounds(x_lim=(x_min, x_max), y_lim=(y_min, y_max))

    if axes_range:
        return patches, axes_range, highlighted_patches, bounds, text_objects

    max_x = max(abs(x_min), abs(x_max))
    max_y = max(abs(y_min), abs(y_max))
    updated_axes_range = max(max_x, max_y) * 1.1

    return patches, updated_axes_range, highlighted_patches, bounds, text_objects


@u.quantity_input(x=u.m, y=u.m, radius=u.m)
def get_telescope_patch(tel_type, x, y, radius, is_grayed_out=False):
    """
    Create patch for a telescope.

    Parameters
    ----------
    tel_type: str
        Telescope type.
    x : Quantity
        X position.
    y : Quantity
        Y position.
    radius : Quantity
        Telescope radius.
    is_grayed_out : bool
        Whether to plot telescope in gray.

    Returns
    -------
    patch : Patch
        Circle or rectangle patch.
    """
    config = leg_h.get_telescope_config(tel_type)
    x, y, r = x.to(u.m), y.to(u.m), radius.to(u.m)

    color = "gray" if is_grayed_out else config["color"]
    fill_flag = True if is_grayed_out else bool(config.get("filled", True))

    if config.get("shape", "circle") == "square":
        return mpatches.Rectangle(
            ((x - r / 2).value, (y - r / 2).value),
            width=r.value,
            height=r.value,
            fill=fill_flag,
            color=color,
        )
    if config.get("shape") == "hexagon":
        return mpatches.RegularPolygon(
            (x.value, y.value),
            numVertices=6,
            radius=r.value * np.sqrt(3) / 2,
            orientation=np.pi / 6,
            fill=fill_flag,
            color=color,
        )

    return mpatches.Circle(
        (x.value, y.value),
        radius=r.value,
        fill=fill_flag,
        alpha=0.5 if is_grayed_out else 1.0,
        color=color,
    )


def get_positions(telescopes):
    """
    Get X/Y positions depending on coordinate system.

    For ground coordinates, rotates the positions by 90 degrees.

    Returns
    -------
    x_rot, y_rot : Quantity
        Position coordinates.
    """
    if "position_x" in telescopes.colnames:
        x, y = telescopes["position_x"], telescopes["position_y"]
        locale_rotate_angle = 90 * u.deg
    elif "utm_east" in telescopes.colnames:
        x, y = telescopes["utm_east"], telescopes["utm_north"]
        locale_rotate_angle = 0 * u.deg
    else:
        raise ValueError("Missing required position columns.")

    return transf.rotate(x, y, locale_rotate_angle) if locale_rotate_angle != 0 else (x, y)


def create_patches(
    telescopes, scale, show_label, ax, grayed_out_elements=None, highlighted_elements=None
):
    """
    Create telescope patches and labels.

    Parameters
    ----------
    telescopes : Table
        Telescope data table.
    scale : float
        Marker size scale factor.
    show_label : bool
        Show telescope labels.
    ax : Axes
        Matplotlib axes object.
    grayed_out_elements : list or None
        List of telescope names to plot as gray circles.
    highlighted_elements : list or None
        List of telescope names to plot with red circles around them.

    Returns
    -------
    patches : list
        Shape patches.
    radii : list
        Telescope radii.
    highlighted_patches : list
        List of highlighted telescope patches.
    text_objects : list
        List of text objects for labels.
    """
    patches, radii, highlighted_patches, text_objects = [], [], [], []
    fontsize, scale_factor = (4, 2) if len(telescopes) > 30 else (8, 1)

    grayed_out_set = set(grayed_out_elements) if grayed_out_elements else set()
    highlighted_set = set(highlighted_elements) if highlighted_elements else set()

    for tel in telescopes:
        name = get_telescope_name(tel)
        radius = get_sphere_radius(tel)
        radii.append(radius)
        try:
            tel_type = names.get_array_element_type_from_name(name)
        except ValueError:
            tel_type = None

        is_grayed_out = name in grayed_out_set
        is_highlighted = name in highlighted_set

        patches.append(
            get_telescope_patch(
                tel_type,
                tel["pos_x_rotated"],
                tel["pos_y_rotated"],
                scale_factor * radius * scale,
                is_grayed_out=is_grayed_out,
            )
        )

        if is_highlighted:
            highlight_patch = mpatches.Circle(
                (tel["pos_x_rotated"].value, tel["pos_y_rotated"].value),
                radius=(scale_factor * radius * scale * 4).value,
                fill=False,
                color="red",
                linewidth=1,
            )
            highlighted_patches.append(highlight_patch)

        if show_label:
            text_objects.append(
                ax.text(
                    tel["pos_x_rotated"].value,
                    tel["pos_y_rotated"].value + scale_factor * radius.value,
                    name,
                    ha="center",
                    va="center",
                    fontsize=fontsize * 0.8,
                )
            )

    return patches, radii, highlighted_patches, text_objects


def get_telescope_name(tel):
    """
    Get telescope name.

    Returns
    -------
    name : str
        Telescope name or fallback identifier.
    """
    if "telescope_name" in tel.colnames:
        return tel["telescope_name"]
    if "asset_code" in tel.colnames and "sequence_number" in tel.colnames:
        return f"{tel['asset_code']}-{tel['sequence_number']}"
    return f"tel_{tel.index}"


def get_sphere_radius(tel):
    """
    Get telescope sphere radius.

    Returns
    -------
    radius : Quantity
        Radius with units.
    """
    return tel["sphere_radius"] if "sphere_radius" in tel.colnames else 10.0 * u.m


def update_legend(ax, telescopes, grayed_out_elements=None, legend_location="best"):
    """Add legend for telescope types and counts."""
    grayed_out_set = set(grayed_out_elements) if grayed_out_elements else set()

    types = []
    for tel in telescopes:
        tel_name = get_telescope_name(tel)
        if tel_name not in grayed_out_set:
            types.append(names.get_array_element_type_from_name(tel_name))

    counts = Counter(types)

    objs, labels = [], []
    handler_map = {}

    for telescope_type in names.get_list_of_array_element_types():
        if counts[telescope_type]:
            objs.append(telescope_type)
            labels.append(f"{telescope_type} ({counts[telescope_type]})")

            class BaseLegendHandlerWrapper:  # pylint: disable=R0903
                """Wrapper for BaseLegendHandler to use in legend."""

                def __init__(self, tel_type):
                    self.tel_type = tel_type

                def legend_artist(self, legend, orig_handle, fontsize, handlebox):
                    handler = leg_h.BaseLegendHandler(self.tel_type)
                    return handler.legend_artist(legend, orig_handle, fontsize, handlebox)

            handler_map[telescope_type] = BaseLegendHandlerWrapper(telescope_type)

    ax.legend(objs, labels, handler_map=handler_map, prop={"size": 11}, loc=legend_location)


def finalize_plot(
    ax,
    patches,
    x_title,
    y_title,
    x_lim=None,
    y_lim=None,
    highlighted_patches=None,
):
    """Finalize plot appearance and limits."""
    ax.add_collection(PatchCollection(patches, match_original=True))

    if highlighted_patches:
        ax.add_collection(PatchCollection(highlighted_patches, match_original=True))

    ax.set(xlabel=x_title, ylabel=y_title)
    ax.tick_params(labelsize=8)
    ax.axis("square")
    if x_lim is not None and y_lim is not None:
        ax.set_xlim(*x_lim)
        ax.set_ylim(*y_lim)
    plt.tight_layout()



#!/usr/bin/python3
"""
    Read array element positions from file and write model parameter files for each element.

    This is an application for experts and should not be used by the general user.
    Reading of input is fine-tuned to the array element files as provided by CTAO.

    Writes one model parameter file per array element into a directory structure compatible
    with the simtools model parameter repository.

    Command line arguments

    input : str
        File containing a table of array element positions.
    repository_path : str
        Path of local copy of model parameter repository.
    parameter_version : str
        Parameter version.
    coordinate_system : str
        Coordinate system of array element positions (ground or utm).

    Examples
    --------
    Add array element positions to repository (ground coordinates):

    .. code-block:: console

        simtools-maintain-simulation-model-write-array-element-positions \
            --input tests/resources/telescope_positions-North-ground.ecsv \
            --output_path /path/to/repository \
            --parameter_version 0.1.0 \
            --coordinate_system ground

    Add array element positions to repository (utm coordinates):

    .. code-block:: console

        simtools-maintain-simulation-model-write-array-element-positions \
            --input tests/resources/telescope_positions-North-utm.ecsv \
            --output_path /path/to/repository \
            --parameter_version 0.1.0 \
            --coordinate_system utm

"""

from simtools.application_control import get_application_label, startup_application
from simtools.configuration import configurator
from simtools.layout.array_layout_utils import write_array_elements_from_file_to_repository


def _parse():
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=get_application_label(__file__),
        description="Add array element positions to model parameter repository",
    )
    config.parser.add_argument(
        "--input",
        help="File containing a table of array element positions.",
        required=False,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system of array element positions (utm or ground).",
        default="ground",
        required=False,
        type=str,
        choices=["ground", "utm"],
    )

    return config.initialize(db_config=True, output=True, simulation_model=["parameter_version"])


def main():
    """Application main."""
    app_context = startup_application(_parse)

    write_array_elements_from_file_to_repository(
        coordinate_system=app_context.args["coordinate_system"],
        input_file=app_context.args["input"],
        repository_path=app_context.args["output_path"],
        parameter_version=app_context.args["parameter_version"],
    )


if __name__ == "__main__":
    main()
